// 
// Copyright (c) 2019 smdn <smdn@smdn.jp>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System.Collections.Generic;

namespace Smdn.Net.MessageAccessProtocols.Strings {
  // required to avoid CS0695
  public abstract class TypedStringEqualityComparerBase : IEqualityComparer<string> {
    private protected TypedStringEqualityComparerBase() { }

    public abstract bool Equals(string x, string y);
    public abstract int GetHashCode(string obj);
  }

  public abstract class TypedStringEqualityComparer<TTypedString> :
    TypedStringEqualityComparerBase /*instead of IEqualityComparer<string>*/,
    IEqualityComparer<TTypedString>
  {
    protected TypedStringEqualityComparer() { }

    internal protected abstract string GetStringExpressionOf(TTypedString typedString);

    public bool Equals(TTypedString x, string y) => Equals(GetStringExpressionOf(x), y);
    public bool Equals(TTypedString x, TTypedString y) => ReferenceEquals(x, y) || Equals(GetStringExpressionOf(x), GetStringExpressionOf(y));
    public int GetHashCode(TTypedString obj) => GetHashCode(GetStringExpressionOf(obj));
  }
}
