//
// Copyright (c) 2010 smdn <smdn@smdn.jp>
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using System.Text;

using Smdn.Net.Pop3.Protocol;

namespace Smdn.Net.Pop3.Client {
  partial class PopMessage {
    /*
     * MarkAsDeletedAsync
     */
    public Task MarkAsDeletedAsync()
    {
      ThrowIfSessionClosed();

      return client.StartTaskExclusively(MarkAsDeletedCore,
                                         CancellationToken.None);
    }

    /*
     * GetUniqueIdAsync
     */
    public Task<string> GetUniqueIdAsync()
    {
      ThrowIfSessionClosed();

      if (isUniqueIdListed) {
        return Task<string>.FromResult(ThrowIfNull(ListedUniqueId));
      }
      else {
        ThrowIfMarkedAsDeleted();

        return client.StartTaskExclusively(() => ThrowIfNull(ListUniqueId()),
                                           CancellationToken.None);
      }
    }

    private static string ThrowIfNull(string uniqueId)
    {
      if (uniqueId == null)
        throw new PopIncapableException(PopCapability.Uidl);
      else
        return uniqueId;
    }

    /*
     * OpenReadAsync
     */
    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<Stream> OpenReadAsync(CancellationToken cancellationToken)
      => Body.OpenReadAsync(cancellationToken: cancellationToken);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<Stream> OpenReadAsync(IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.OpenReadAsync(progress: progress);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<Stream> OpenReadAsync(CancellationToken cancellationToken,
                                      IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.OpenReadAsync(cancellationToken: cancellationToken, progress: progress);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<Stream> OpenReadAsync(int maxLines,
                                      IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.OpenReadAsync(maxLines: maxLines, progress: progress);

    /*
     * ReadAsAsync(Converter<Stream>)
     */
    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<TResult> ReadAsAsync<TResult>(Converter<Stream, TResult> converter)
      => Body.ReadAsAsync(converter: ReadAsAsyncConverter, arg: converter ?? throw new ArgumentNullException(nameof(converter)));

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<TResult> ReadAsAsync<TResult>(Converter<Stream, TResult> converter,
                                              CancellationToken cancellationToken)
      => Body.ReadAsAsync(converter: ReadAsAsyncConverter, arg: converter ?? throw new ArgumentNullException(nameof(converter)), cancellationToken: cancellationToken);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<TResult> ReadAsAsync<TResult>(Converter<Stream, TResult> converter,
                                              IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.ReadAsAsync(converter: ReadAsAsyncConverter, arg: converter ?? throw new ArgumentNullException(nameof(converter)), progress: progress);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<TResult> ReadAsAsync<TResult>(Converter<Stream, TResult> converter,
                                              CancellationToken cancellationToken,
                                              IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.ReadAsAsync(converter: ReadAsAsyncConverter, arg: converter ?? throw new ArgumentNullException(nameof(converter)), cancellationToken: cancellationToken, progress: progress);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<TResult> ReadAsAsync<TResult>(int maxLines,
                                              Converter<Stream, TResult> converter)
      => Body.ReadAsAsync(converter: ReadAsAsyncConverter, arg: converter ?? throw new ArgumentNullException(nameof(converter)), maxLines: maxLines);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<TResult> ReadAsAsync<TResult>(int maxLines,
                                              Converter<Stream, TResult> converter,
                                              CancellationToken cancellationToken)
      => Body.ReadAsAsync(converter: ReadAsAsyncConverter, arg: converter ?? throw new ArgumentNullException(nameof(converter)), maxLines: maxLines, cancellationToken: cancellationToken);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<TResult> ReadAsAsync<TResult>(int maxLines,
                                              Converter<Stream, TResult> converter,
                                              IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.ReadAsAsync(converter: ReadAsAsyncConverter, arg: converter ?? throw new ArgumentNullException(nameof(converter)), maxLines: maxLines, progress: progress);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<TResult> ReadAsAsync<TResult>(int maxLines,
                                              Converter<Stream, TResult> converter,
                                              CancellationToken cancellationToken,
                                              IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.ReadAsAsync(converter: ReadAsAsyncConverter, arg: converter ?? throw new ArgumentNullException(nameof(converter)), maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    [Obsolete]
    private static Task<TResult> ReadAsAsyncConverter<TResult>(Stream stream, Converter<Stream, TResult> converter)
      => Task.FromResult(converter(stream));

    /*
     * ReadAsAsync<TResult>(Func<Stream, ...>)
     */

    /*
     * ReadAllBytesAsync()
     */
    [Obsolete("use ReadAsByteArrayAsync() instead")]
    public Task<byte[]> ReadAllBytesAsync()
      => Body.ReadAsByteArrayAsync();

    [Obsolete("use ReadAsByteArrayAsync() instead")]
    public Task<byte[]> ReadAllBytesAsync(CancellationToken cancellationToken)
      => Body.ReadAsByteArrayAsync(cancellationToken: cancellationToken);

    [Obsolete("use ReadAsByteArrayAsync() instead")]
    public Task<byte[]> ReadAllBytesAsync(IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.ReadAsByteArrayAsync(progress: progress);

    [Obsolete("use ReadAsByteArrayAsync() instead")]
    public Task<byte[]> ReadAllBytesAsync(CancellationToken cancellationToken,
                                          IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.ReadAsByteArrayAsync(cancellationToken: cancellationToken, progress: progress);

    [Obsolete("use ReadAsByteArrayAsync() instead")]
    public Task<byte[]> ReadAllBytesAsync(int maxLines)
      => Body.ReadAsByteArrayAsync(maxLines: maxLines);

    [Obsolete("use ReadAsByteArrayAsync() instead")]
    public Task<byte[]> ReadAllBytesAsync(int maxLines,
                                          CancellationToken cancellationToken)
      => Body.ReadAsByteArrayAsync(maxLines: maxLines, cancellationToken: cancellationToken);

    [Obsolete("use ReadAsByteArrayAsync() instead")]
    public Task<byte[]> ReadAllBytesAsync(int maxLines,
                                          IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.ReadAsByteArrayAsync(maxLines: maxLines, progress: progress);

    [Obsolete("use ReadAsByteArrayAsync() instead")]
    public Task<byte[]> ReadAllBytesAsync(int maxLines,
                                          CancellationToken cancellationToken,
                                          IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.ReadAsByteArrayAsync(maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    /*
     * WriteToAsync(Stream)
     */
    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task WriteToAsync(Stream stream,
                             CancellationToken cancellationToken)
      => Body.WriteToAsync(stream: stream, cancellationToken: cancellationToken);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task WriteToAsync(Stream stream,
                             IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.WriteToAsync(stream: stream, progress: progress);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task WriteToAsync(Stream stream,
                             CancellationToken cancellationToken,
                             IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.WriteToAsync(stream: stream, cancellationToken: cancellationToken, progress: progress);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task WriteToAsync(Stream stream,
                             int maxLines,
                             IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.WriteToAsync(stream: stream, maxLines: maxLines, progress: progress);

    /*
     * WriteToAsync(BinaryWriter)
     */
    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task WriteToAsync(BinaryWriter writer,
                             CancellationToken cancellationToken)
      => Body.WriteToAsync(writer: writer, cancellationToken: cancellationToken);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task WriteToAsync(BinaryWriter writer,
                             IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.WriteToAsync(writer: writer, progress: progress);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task WriteToAsync(BinaryWriter writer,
                             CancellationToken cancellationToken,
                             IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.WriteToAsync(writer: writer, cancellationToken: cancellationToken, progress: progress);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task WriteToAsync(BinaryWriter writer,
                             int maxLines,
                             IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.WriteToAsync(writer: writer, maxLines: maxLines, progress: progress);

    /*
     * SaveAsync()
     */
    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task SaveAsync(string path,
                          CancellationToken cancellationToken)
      => Body.SaveAsync(path: path, cancellationToken: cancellationToken);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task SaveAsync(string path,
                          IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.SaveAsync(path: path, progress: progress);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task SaveAsync(string path,
                          CancellationToken cancellationToken,
                          IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.SaveAsync(path: path, cancellationToken: cancellationToken, progress: progress);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task SaveAsync(string path,
                          int maxLines,
                          IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.SaveAsync(path: path, maxLines: maxLines, progress: progress);

    /*
     * OpenTextAsync()
     */
    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<StreamReader> OpenTextAsync(int maxLines)
      => Body.OpenTextAsync(maxLines: maxLines);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<StreamReader> OpenTextAsync(Encoding charset,
                                            int maxLines,
                                            IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.OpenTextAsync(charset: charset ?? throw new ArgumentNullException(nameof(charset)), maxLines: maxLines, progress: progress);

    /*
     * ReadAsAsync<TResult>(Converter<StreamReader>)
     */
    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<TResult> ReadAsAsync<TResult>(Converter<StreamReader, TResult> converter)
      => Body.ReadAsAsync(converter: ReadAsAsyncConverter, arg: converter ?? throw new ArgumentNullException(nameof(converter)));

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<TResult> ReadAsAsync<TResult>(Encoding charset,
                                              Converter<StreamReader, TResult> converter)
      => Body.ReadAsAsync(converter: ReadAsAsyncConverter, arg: converter ?? throw new ArgumentNullException(nameof(converter)), charset: charset ?? throw new ArgumentNullException(nameof(charset)));

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<TResult> ReadAsAsync<TResult>(int maxLines,
                                              Converter<StreamReader, TResult> converter)
      => Body.ReadAsAsync(converter: ReadAsAsyncConverter, arg: converter ?? throw new ArgumentNullException(nameof(converter)), maxLines: maxLines);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<TResult> ReadAsAsync<TResult>(Encoding charset,
                                              int maxLines,
                                              Converter<StreamReader, TResult> converter)
      => Body.ReadAsAsync(converter: ReadAsAsyncConverter, arg: converter ?? throw new ArgumentNullException(nameof(converter)), charset: charset ?? throw new ArgumentNullException(nameof(charset)), maxLines: maxLines);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<TResult> ReadAsAsync<TResult>(Encoding charset,
                                              int maxLines,
                                              Converter<StreamReader, TResult> converter,
                                              CancellationToken cancellationToken)
      => Body.ReadAsAsync(converter: ReadAsAsyncConverter, arg: converter ?? throw new ArgumentNullException(nameof(converter)), charset: charset ?? throw new ArgumentNullException(nameof(charset)), maxLines: maxLines, cancellationToken: cancellationToken);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<TResult> ReadAsAsync<TResult>(Encoding charset,
                                              int maxLines,
                                              Converter<StreamReader, TResult> converter,
                                              IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.ReadAsAsync(converter: ReadAsAsyncConverter, arg: converter ?? throw new ArgumentNullException(nameof(converter)), charset: charset ?? throw new ArgumentNullException(nameof(charset)), maxLines: maxLines, progress: progress);

    [Obsolete(PopClient.ObsoleteAttributeMessage_UseOptionalParameterOverloads)]
    public Task<TResult> ReadAsAsync<TResult>(Encoding charset,
                                              int maxLines,
                                              Converter<StreamReader, TResult> converter,
                                              CancellationToken cancellationToken,
                                              IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.ReadAsAsync(converter: ReadAsAsyncConverter, arg: converter ?? throw new ArgumentNullException(nameof(converter)), charset: charset ?? throw new ArgumentNullException(nameof(charset)), maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);

    [Obsolete]
    private static Task<TResult> ReadAsAsyncConverter<TResult>(StreamReader reader, Converter<StreamReader, TResult> converter)
      => Task.FromResult(converter(reader));

    /*
     * ReadLinesAsync()
     */

    /*
    * ReadAllLinesAsync()
    */
    [Obsolete("use ReadAsStringLinesAsync() instead")]
    public Task<string[]> ReadAllLinesAsync()
      => ToArrayAsync(Body.ReadAsStringLinesAsync());

    [Obsolete("use ReadAsStringLinesAsync() instead")]
    public Task<string[]> ReadAllLinesAsync(Encoding charset)
      => ToArrayAsync(Body.ReadAsStringLinesAsync(charset: charset ?? throw new ArgumentNullException(nameof(charset))));

    [Obsolete("use ReadAsStringLinesAsync() instead")]
    public Task<string[]> ReadAllLinesAsync(int maxLines)
      => ToArrayAsync(Body.ReadAsStringLinesAsync(maxLines: maxLines));

    [Obsolete("use ReadAsStringLinesAsync() instead")]
    public Task<string[]> ReadAllLinesAsync(Encoding charset,
                                            int maxLines)
      => ToArrayAsync(Body.ReadAsStringLinesAsync(charset: charset ?? throw new ArgumentNullException(nameof(charset)), maxLines: maxLines));

    [Obsolete("use ReadAsStringLinesAsync() instead")]
    public Task<string[]> ReadAllLinesAsync(Encoding charset,
                                            int maxLines,
                                            CancellationToken cancellationToken)
      => ToArrayAsync(Body.ReadAsStringLinesAsync(charset: charset ?? throw new ArgumentNullException(nameof(charset)), maxLines: maxLines, cancellationToken: cancellationToken));

    [Obsolete("use ReadAsStringLinesAsync() instead")]
    public Task<string[]> ReadAllLinesAsync(Encoding charset,
                                            int maxLines,
                                            IProgress<PopMessageRetrieveProgressInfo> progress)
      => ToArrayAsync(Body.ReadAsStringLinesAsync(charset: charset ?? throw new ArgumentNullException(nameof(charset)), maxLines: maxLines, progress: progress));

    [Obsolete("use ReadAsStringLinesAsync() instead")]
    public Task<string[]> ReadAllLinesAsync(Encoding charset,
                                            int maxLines,
                                            CancellationToken cancellationToken,
                                            IProgress<PopMessageRetrieveProgressInfo> progress)
      => ToArrayAsync(Body.ReadAsStringLinesAsync(charset: charset ?? throw new ArgumentNullException(nameof(charset)), maxLines: maxLines, cancellationToken: cancellationToken, progress: progress));

    private static async Task<string[]> ToArrayAsync(Task<IReadOnlyList<string>> task)
      => (await task.ConfigureAwait(false)).ToArray();

    /*
     * ReadAllTextAsync()
     */
    [Obsolete("use ReadAsStringAsync() instead")]
    public Task<string> ReadAllTextAsync()
      => Body.ReadAsStringAsync();

    [Obsolete("use ReadAsStringAsync() instead")]
    public Task<string> ReadAllTextAsync(Encoding charset)
      => Body.ReadAsStringAsync(charset: charset ?? throw new ArgumentNullException(nameof(charset)));

    [Obsolete("use ReadAsStringAsync() instead")]
    public Task<string> ReadAllTextAsync(int maxLines)
      => Body.ReadAsStringAsync(maxLines: maxLines);

    [Obsolete("use ReadAsStringAsync() instead")]
    public Task<string> ReadAllTextAsync(Encoding charset,
                                         int maxLines)
      => Body.ReadAsStringAsync(charset: charset ?? throw new ArgumentNullException(nameof(charset)), maxLines: maxLines);

    [Obsolete("use ReadAsStringAsync() instead")]
    public Task<string> ReadAllTextAsync(Encoding charset,
                                         int maxLines,
                                         CancellationToken cancellationToken)
      => Body.ReadAsStringAsync(charset: charset ?? throw new ArgumentNullException(nameof(charset)), maxLines: maxLines, cancellationToken: cancellationToken);

    [Obsolete("use ReadAsStringAsync() instead")]
    public Task<string> ReadAllTextAsync(Encoding charset,
                                         int maxLines,
                                         IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.ReadAsStringAsync(charset: charset ?? throw new ArgumentNullException(nameof(charset)), maxLines: maxLines, progress: progress);

    [Obsolete("use ReadAsStringAsync() instead")]
    public Task<string> ReadAllTextAsync(Encoding charset,
                                         int maxLines,
                                         CancellationToken cancellationToken,
                                         IProgress<PopMessageRetrieveProgressInfo> progress)
      => Body.ReadAsStringAsync(charset: charset ?? throw new ArgumentNullException(nameof(charset)), maxLines: maxLines, cancellationToken: cancellationToken, progress: progress);
  }
}
