using System;
using System.Collections.Generic;
using System.Linq;

using Smdn.Net.MessageAccessProtocols.Strings;
using Smdn.Net.Pop3.Protocol.Strings;

namespace Smdn.Net.Pop3.Protocol {
  internal abstract class PopPseudoEnumValueSet<TEnumValue> : HashSet<TEnumValue>, IReadOnlyStringSet<TEnumValue> where TEnumValue : PopEnumValue {
    private class EnumValueEqualityComparer : EqualityComparer<TEnumValue> {
      public override bool Equals(TEnumValue x, TEnumValue y) => StringComparer.OrdinalIgnoreCase.Equals(x?.Value, y?.Value);
      public override int GetHashCode(TEnumValue obj) => StringComparer.OrdinalIgnoreCase.GetHashCode(obj?.Value);
    }

    protected PopPseudoEnumValueSet(IEnumerable<TEnumValue> items)
      : base(items, new EnumValueEqualityComparer())
    {
    }

    public bool Contains(string value) => this.Any(i => Equals(value, i));
    public bool TryGetValue(string value, out TEnumValue item) => (item = this.FirstOrDefault(i => Equals(value, i))) != null;
    private bool Equals(string value, TEnumValue item) => StringComparer.OrdinalIgnoreCase.Equals(value, item?.Value);
  }

  internal class PopPseudoCapabilitySet : PopPseudoEnumValueSet<PopCapability> {
    public static readonly PopPseudoCapabilitySet Empty = new PopPseudoCapabilitySet();

    public PopPseudoCapabilitySet(params PopCapability[] capabilities)
      : base(capabilities)
    {
    }
  }
}
